import { ApiException, ServiceException, ParamsException } from './exception';
import * as utils from './utils';
import _ from 'lodash';

import {
  ApiResKeyMap,
  ExecuteHttpOpt,
  HttpExecuteRecord,
  OnSuccess,
  OnServiceException,
  OnApiException,
  OnParamsException,
  BeforeExecuteHttp,
  AfterExecuteHttp,
  BeforeExecuteMain,
  AfterExecuteMain,
} from './types';
import { AxiosRequestConfig, AxiosInstance, Method, AxiosPromise } from 'axios';
import { Schema } from 'joi';

export interface AppConfig {
  axios: AxiosInstance;
  apiResKeyMap?: ApiResKeyMap;
  onSuccess?: OnSuccess;
  onServiceException?: OnServiceException;
  onApiException?: OnApiException;
  onParamsException?: OnParamsException;
  beforeExecuteHttp?: BeforeExecuteHttp;
  afterExecuteHttp?: AfterExecuteHttp;
  beforeExecuteMain?: BeforeExecuteMain;
  afterExecuteMain?: AfterExecuteMain;
}

export class PagodaServerless {
  httpRequestKey: number;

  apiResKeyMap: ApiResKeyMap;

  axios: AxiosInstance;

  httpExecuteStack: HttpExecuteRecord[];

  readonly utils: typeof utils;

  readonly ParamsException = ParamsException;

  readonly ApiException = ApiException;

  readonly ServiceException = ServiceException;

  onSuccess?: OnSuccess;

  onServiceException?: OnServiceException;

  onApiException?: OnApiException;

  onParamsException?: OnParamsException;

  beforeExecuteHttp?: BeforeExecuteHttp;

  afterExecuteHttp?: AfterExecuteHttp;

  beforeExecuteMain?: BeforeExecuteMain;

  afterExecuteMain?: AfterExecuteMain;

  constructor(config: AppConfig) {
    this.httpRequestKey = 1;
    this.httpExecuteStack = [];
    this.apiResKeyMap = {
      middle: {
        resSuccessCode: 0,
        resCodeKey: 'resultCode',
        resMessageKey: 'errorMsg',
        resDataKey: 'data',
        desc: '中台',
      },
      ...config.apiResKeyMap,
    };
    this.axios = config.axios;
    this.utils = utils;
    this.onSuccess = config.onSuccess;
    this.onServiceException = config.onServiceException;
    this.onApiException = config.onApiException;
    this.onParamsException = config.onParamsException;
    this.beforeExecuteHttp = config.beforeExecuteHttp;
    this.afterExecuteHttp = config.afterExecuteHttp;
    this.beforeExecuteMain = config.beforeExecuteMain;
    this.afterExecuteMain = config.afterExecuteMain;
  }

  private http(
    method: Method,
    url: string,
    data: Record<string, any> = {},
    config: AxiosRequestConfig | undefined
  ) {
    const axiosConfig = {
      ...config,
      method,
      url,
    };

    if (method.toLowerCase() === 'get') {
      axiosConfig.params = data;
    } else {
      axiosConfig.data = data;
    }

    return this.axios(axiosConfig);
  }

  private async execute({
    apiPromise,
    origin = '未知',
    resCodeKey,
    resMessageKey,
    resDataKey,
    resSuccessCode,
    pretreatment = true,
    ignoreErrorCode = false,
    errorCode,
    errorMsg,
    errorExtraInfo,
  }: {
    apiPromise: AxiosPromise<any>;
    origin: string;
    resCodeKey: string;
    resMessageKey: string;
    resDataKey: string | undefined;
    resSuccessCode: any;
    pretreatment: boolean;
    ignoreErrorCode: boolean;
    errorCode?: number;
    errorMsg?: string;
    errorExtraInfo?: any;
  }) {
    const record = this.startRecord();

    this.beforeExecuteHttp &&
      this.beforeExecuteHttp({
        key: record.key,
        origin,
      });

    let reqRes;

    try {
      reqRes = (await apiPromise).data;
    } catch (e) {
      const { response, config } = e;

      throw new ServiceException({
        message: e.message,
        origin,
        extraInfo: response
          ? {
              originResponse: response.data,
              config,
            }
          : null,
      });
    }

    const { [resCodeKey]: code, [resMessageKey]: msg } = reqRes;

    this.endRecord(origin, record);

    this.afterExecuteHttp &&
      this.afterExecuteHttp({
        key: record.key,
        origin,
      });

    if (!ignoreErrorCode && code !== resSuccessCode) {
      throw new ApiException({
        code: errorCode || code,
        message: errorMsg || msg,
        origin,
        extraInfo: errorExtraInfo,
      });
    }

    return resDataKey && pretreatment ? reqRes[resDataKey] : reqRes;
  }

  private startRecord() {
    const key = this.httpRequestKey++;
    const start = Date.now();

    return {
      key,
      start,
    };
  }

  private endRecord(
    origin: string,
    {
      key,
      start,
    }: {
      key: number;
      start: number;
    }
  ) {
    const end = Date.now();
    this.httpExecuteStack.push({
      key,
      origin,
      time: (end - start) / 1000,
    });
  }

  private getHttpExecuteStack(): HttpExecuteRecord[] {
    return this.httpExecuteStack.sort((a, b) => a.key - b.key);
  }

  validateParamsByJoi({ schema, data }: { schema: Schema; data: any }) {
    const { error } = schema.validate(data);

    if (error) {
      const [errorDetail] = error.details;
      const prop = errorDetail.context?.label;

      if (prop) {
        throw new ParamsException({
          prop,
          message: `接口${prop}参数异常`,
          isJoi: true,
          joiErrorDetail: errorDetail,
        });
      } else {
        throw new ParamsException({
          message: `接口参数异常`,
          isJoi: true,
          joiErrorDetail: errorDetail,
        });
      }
    }
  }

  checkParamsEmpty(params: any, props?: string[]) {
    if (props && _.isObject(params)) {
      props.forEach((prop) => {
        const value = (params as Record<string, any>)[prop];

        if (this.utils.isEmpty(value)) {
          throw new ParamsException({
            prop,
            message: `缺少参数${prop}字段`,
          });
        }
      });
    } else if (this.utils.isEmpty(params)) {
      throw new ParamsException({
        message: `缺少接口入参`,
      });
    }
  }

  // http 调用 API
  async executeHttp(opt: ExecuteHttpOpt) {
    const {
      method = 'post',
      url = '',
      data,
      config,
      type = 'middle',
      apiPromise = this.http(method, url, data, config),
      apiDesc,
      pretreatment = true,
      ignoreErrorCode = false,
      errorMsg,
      errorCode,
    } = opt;

    if (!this.apiResKeyMap[type]) {
      throw new ServiceException({
        origin: 'pagodaServerless executeHttp函数',
        message: `${type}接口类型不存在`,
      });
    }

    const {
      resSuccessCode,
      resCodeKey,
      resMessageKey,
      resDataKey,
      desc,
    } = this.apiResKeyMap[type];

    const result = await this.execute({
      apiPromise,
      origin: `${desc}API：${apiDesc || url}`,
      resCodeKey,
      resMessageKey,
      resDataKey,
      resSuccessCode,
      pretreatment,
      ignoreErrorCode,
      errorMsg,
      errorCode,
      errorExtraInfo: {
        url,
        method,
        data,
        config,
      },
    });

    return result;
  }

  // 注入方式调用中台 API
  async executeApi(apiPromise: Promise<any>, config?: ExecuteHttpOpt) {
    const data = await this.executeHttp({
      ...config,
      apiDesc: config?.apiDesc || '未知',
      apiPromise,
    });

    return data;
  }

  async executeMain(
    main: () => Promise<any>,
    config: { returnRes: boolean } = {
      returnRes: true,
    }
  ) {
    const { returnRes } = config;

    try {
      this.beforeExecuteMain && this.beforeExecuteMain();

      const data = await main();

      this.afterExecuteMain && this.afterExecuteMain();

      if (returnRes) {
        this.onSuccess && this.onSuccess(data, this.getHttpExecuteStack());
      }
    } catch (e) {
      const httpExecuteStack = this.getHttpExecuteStack();

      if (e.constructor === ApiException) {
        this.onApiException && this.onApiException(e, httpExecuteStack);
      } else if (e.constructor === ParamsException) {
        this.onParamsException && this.onParamsException(e, httpExecuteStack);
      } else if (this.onServiceException) {
        const err =
          e.constructor === ServiceException
            ? e
            : new ServiceException({
                origin: e.stack,
                message: e.message,
              });

        this.onServiceException(err, httpExecuteStack);
      }
    }
  }
}
